from simtk.openmm.app import *
from simtk.openmm import *
from simtk.unit import *
from sys import stdout
import pandas
import numpy as np

pdb = PDBFile('start.pdb')
forcefield = ForceField('ff.xml')
system = forcefield.createSystem(pdb.topology,)
system.setDefaultPeriodicBoxVectors([100,0,0],[0,100,0],[0,0,100])

cols=['atom_index','atom_id','atom_name',
      'residue_index','residue_id','residue_name',
      'chain_index','chain_id']
atom_list=pandas.DataFrame(columns=cols)
for residue in pdb.topology.residues():
    for atom in residue.atoms():
        atom_list.loc[atom.index]=[atom.index,atom.id,atom.name,
                                   residue.index,residue.id,residue.name,
                                   residue.chain.index,residue.chain.id]
w= [2.45183638, -1.87012941, -0.58845876]
w2=[0.79186638, -0.64395771, -0.37867282]
w3=[1.66588279, -1.26981613, -0.12620828]

cw1=[-0.36157599,  0.38981825,  0.0549247]
cw2=[ 0.27684821, -0.30509075,  0.0549247]
cw3=[-0.30509096,  0.27684879, -0.0549247]
cw4=[ 0.38981873, -0.36157629, -0.0549247]

for _,res in atom_list.groupby(['chain_index','residue_id']):
    assert len(res['residue_name'].unique())==1
    resname=res['residue_name'].unique()[0]
    if resname=='ACT':
        a2=int(res[res['atom_name']=='A2'].index[0])
        a1=int(res[res['atom_name']=='A1'].index[0])
        a3=int(res[res['atom_name']=='A3'].index[0])
        a5=int(res[res['atom_name']=='A5'].index[0])
        virtual_site=OutOfPlaneSite(a2,a1,a3,w[0],w[1],w[2])
        system.setVirtualSite(a5,virtual_site)
        a6=int(res[res['atom_name']=='A6'].index[0])
        virtual_site=OutOfPlaneSite(a2,a1,a3,w2[0],w2[1],w2[2])
        system.setVirtualSite(a6,virtual_site)
        a7=int(res[res['atom_name']=='A7'].index[0])
        virtual_site=OutOfPlaneSite(a2,a1,a3,w3[0],w3[1],w3[2])
        system.setVirtualSite(a7,virtual_site)
    if resname=='CAM':
        c1=int(res[res['atom_name']=='Cx1'].index[0])
        c2=int(res[res['atom_name']=='Cx2'].index[0])
        c3=int(res[res['atom_name']=='Cx3'].index[0])
        c01=OutOfPlaneSite(c1,c2,c3,cw1[0],cw1[1],cw1[2])
        c02=OutOfPlaneSite(c1,c2,c3,cw2[0],cw2[1],cw2[2])
        c07=OutOfPlaneSite(c1,c2,c3,cw3[0],cw3[1],cw3[2])
        c08=OutOfPlaneSite(c1,c2,c3,cw4[0],cw4[1],cw4[2])
        c03=OutOfPlaneSite(c2,c3,c1,cw1[0],cw1[1],cw1[2])
        c04=OutOfPlaneSite(c2,c3,c1,cw2[0],cw2[1],cw2[2])
        c09=OutOfPlaneSite(c2,c3,c1,cw3[0],cw3[1],cw3[2])
        c10=OutOfPlaneSite(c2,c3,c1,cw4[0],cw4[1],cw4[2])
        c05=OutOfPlaneSite(c3,c1,c2,cw1[0],cw1[1],cw1[2])
        c06=OutOfPlaneSite(c3,c1,c2,cw2[0],cw2[1],cw2[2])
        c11=OutOfPlaneSite(c3,c1,c2,cw3[0],cw3[1],cw3[2])
        c12=OutOfPlaneSite(c3,c1,c2,cw4[0],cw4[1],cw4[2])
        system.setVirtualSite(int(res[res['atom_name']=='C01'].index[0]),c01)
        system.setVirtualSite(int(res[res['atom_name']=='C02'].index[0]),c02)
        system.setVirtualSite(int(res[res['atom_name']=='C03'].index[0]),c03)
        system.setVirtualSite(int(res[res['atom_name']=='C04'].index[0]),c04)
        system.setVirtualSite(int(res[res['atom_name']=='C05'].index[0]),c05)
        system.setVirtualSite(int(res[res['atom_name']=='C06'].index[0]),c06)
        system.setVirtualSite(int(res[res['atom_name']=='C07'].index[0]),c07)
        system.setVirtualSite(int(res[res['atom_name']=='C08'].index[0]),c08)
        system.setVirtualSite(int(res[res['atom_name']=='C09'].index[0]),c09)
        system.setVirtualSite(int(res[res['atom_name']=='C10'].index[0]),c10)
        system.setVirtualSite(int(res[res['atom_name']=='C11'].index[0]),c11)
        system.setVirtualSite(int(res[res['atom_name']=='C12'].index[0]),c12)
        cc=ThreeParticleAverageSite(c1,c2,c3,1/3.,1/3.,1/3.)
        system.setVirtualSite(int(res[res['atom_name']=='Cc'].index[0]),cc)
atom_list['Virtual']=[system.isVirtualSite(a) for a in range(len(atom_list))]

data=[['A1', 'A2', 25.85, 4.12 , 0],
      ['A1', 'A3', 25.77, 7.03, 0],
      ['A3', 'A4', 24.67, 8.64, 0],
      ['A2', 'A3', 29.41, 0.84, 1],
      ['A2', 'A4', 32.65, 0.53, 1],
      ['A4', 'A1', 33.27, 1.1, 1],
      ['A4', 'A3', 31.46, 0.4, 1],
      ['A3', 'A3', 41.06, 0.99, 1],
      ['A2', 'A1', 35.18, 0.02, 2],
      ['A2', 'A3', 28.61, 0.3, 2],
      ['A4', 'A3', 37.44, 0.89, 2],
      ['A1', 'A1', 57.88, 0.67, 2],
      ['A2', 'A2', 57.09, 0.44, 2],
      ['A4', 'A4', 56.93, 0.63, 2]]
actin_bonds=pandas.DataFrame(data,columns=['i','j','x0','k','skip'])
actin_bonds['type']='1'+actin_bonds['i']+'-'+(1+actin_bonds['skip']).astype(str)+actin_bonds['j']
actin_bonds

data=[['A1', 'A3','A4', 91.78, 926.62],
      ['A2', 'A1','A3', 94.93, 599.12]]
actin_angles=pandas.DataFrame(data,columns=['i','j','l','x0','k'])
actin_angles['type']=actin_angles['i']+'-'+actin_angles['j']+'-'+actin_angles['l']
actin_angles

data=[['A2','A1', 'A3','A4',1, 155.89, 477.25]]
actin_dihedrals=pandas.DataFrame(data,columns=['i','j','l','m','periodicity','phase','k'])
actin_dihedrals['type']=actin_dihedrals['i']+'-'+actin_dihedrals['j']+'-'+actin_dihedrals['l']+'-'+actin_dihedrals['m']
actin_dihedrals

#Bonds and angles
sbonds = pandas.DataFrame(columns=['i','j','x0','k','type'])
sangles = pandas.DataFrame(columns=['i','j','l','x0','k','type'])
sdihedrals = pandas.DataFrame(columns=['i','j','l','m','periodicity','phase','k','type'])
for mol in atom_list['chain_index'].unique():
    sel=atom_list[atom_list['chain_index']==mol]
    name=sel['residue_name'].unique()
    assert(len(name==0))
    name=name[0]
    if name in ['ACT','ACD']:
        for _,b in actin_bonds.iterrows():
            i1=sel[sel['atom_name']==b['i']].index[b['skip']:]
            if b['skip']==0:
                i2=sel[sel['atom_name']==b['j']].index[:]
            else:
                i2=sel[sel['atom_name']==b['j']].index[:-b['skip']]
            assert(len(i1)==len(i2))
            for i,j in zip(i1,i2):
                sbonds.loc[len(sbonds)]=[i,j,b['x0'],b['k'],b['type']]
        for _,a in actin_angles.iterrows():
            i1=sel[sel['atom_name']==a['i']].index
            i2=sel[sel['atom_name']==a['j']].index
            i3=sel[sel['atom_name']==a['l']].index
            assert(len(i1)==len(i2)==len(i3))
            for i,j,l in zip(i1,i2,i3):
                sangles.loc[len(sangles)]=[i,j,l,a['x0'],a['k'],a['type']]
        for _,d in actin_dihedrals.iterrows():
            i1=sel[sel['atom_name']==d['i']].index
            i2=sel[sel['atom_name']==d['j']].index
            i3=sel[sel['atom_name']==d['l']].index
            i4=sel[sel['atom_name']==d['m']].index
            assert(len(i1)==len(i2)==len(i3)==len(i4))
            for i,j,l,m in zip(i1,i2,i3,i4):
                sdihedrals.loc[len(sdihedrals)]=[i,j,l,m,d['periodicity'],d['phase'],d['k'],d['type']]

    if name=='CAM':
        for res in sel['residue_index'].unique():
            ax=sel[(sel['residue_index']==res) & sel['atom_name'].isin(['Cx1','Cx2','Cx3'])].index
            dl=pandas.DataFrame([ax[[0,1,2]],
                                 ax[[1,2,0]]],
                                index=['i','j']).T
            dl['x0']=49.5682
            dl['k']=10
            dl['type']='hex'

            sbonds=pandas.concat([sbonds,dl])
sbonds=sbonds.reset_index(drop=True)

#Non-bonded force
data=[['A2','A2',10,5.600,7.,"epsilon*(sigma/r)^12"],
      ['Cc','Cc',10,8.000,12.,"epsilon*(sigma/r)^12"],
      ['A2','Cc',10,5.800,7.,"epsilon*(sigma/r)^12"]]
#data=[]
repulsion=pandas.DataFrame(data,columns=['i','j','epsilon','sigma','cutoff','force'])


[system.removeForce(0) for i,f in enumerate(system.getForces())]
harmonic_bond = HarmonicBondForce()
harmonic_bond.setUsesPeriodicBoundaryConditions(True)
for i,b in sbonds.iterrows():
    harmonic_bond.addBond(int(b['i']),int(b['j']),b['x0']/10,b['k']*4.184*100)
system.addForce(harmonic_bond)

harmonic_angle = HarmonicAngleForce()
harmonic_angle.setUsesPeriodicBoundaryConditions(True)
for i,b in sangles.iterrows():
    harmonic_angle.addAngle(int(b['i']),int(b['j']),int(b['l']),b['x0']/180*np.pi,b['k']*4.184)
system.addForce(harmonic_angle)

harmonic_torsion = PeriodicTorsionForce()
harmonic_torsion.setUsesPeriodicBoundaryConditions(True)
for i,b in sdihedrals.iterrows():
    harmonic_torsion.addTorsion(int(b['i']),int(b['j']),int(b['l']),int(b['m']),b['periodicity'],b['phase']/180*np.pi,b['k']*4.184)
system.addForce(harmonic_torsion)

for _,r in repulsion.iterrows():
    rf=CustomNonbondedForce(r['force'])
    rf.setNonbondedMethod(rf.CutoffPeriodic)
    rf.addGlobalParameter('epsilon',r['epsilon'])
    rf.addGlobalParameter('sigma',r['sigma'])
    rf.setCutoffDistance(r['cutoff'])
    rf.setUseLongRangeCorrection(False)
    for _,a in atom_list.iterrows():
        rf.addParticle()
    sel1=atom_list[atom_list['atom_name']==r['i']]
    sel2=atom_list[atom_list['atom_name']==r['j']]
    rf.addInteractionGroup(sel1.index,sel2.index)
    rf.createExclusionsFromBonds(sbonds[['i','j']].values,2)
    system.addForce(rf)


#CustomHBond
gaussian=CustomHbondForce("-epsi*g1;"
                          "g1=(exp(-10*w*dd)+exp(-100*w*dd))/2;"
                          "dd=(dist1^2);"
                          "dist1= distance(a1,d1);")

gaussian.addGlobalParameter('epsi',100) #Energy minimum
gaussian.addGlobalParameter('w',0.02) #well width
gaussian.setNonbondedMethod(gaussian.CutoffPeriodic)
gaussian.setCutoffDistance(12)


#Aceptors
A1=atom_list[atom_list['atom_name']=='A5'].index
A2=atom_list[atom_list['atom_name']=='A6'].index
A3=atom_list[atom_list['atom_name']=='A7'].index
assert len(A1)==len(A2)==len(A3)
for a1,a2,a3 in zip(A1,A2,A3):
    gaussian.addAcceptor(a1,a2,a3)

#Donors
Cm=[atom_list[atom_list['atom_name']=='C01'].index,
    atom_list[atom_list['atom_name']=='C02'].index,
    atom_list[atom_list['atom_name']=='C03'].index,
    atom_list[atom_list['atom_name']=='C04'].index,
    atom_list[atom_list['atom_name']=='C05'].index,
    atom_list[atom_list['atom_name']=='C06'].index,
    atom_list[atom_list['atom_name']=='C07'].index,
    atom_list[atom_list['atom_name']=='C08'].index,
    atom_list[atom_list['atom_name']=='C09'].index,
    atom_list[atom_list['atom_name']=='C10'].index,
    atom_list[atom_list['atom_name']=='C11'].index,
    atom_list[atom_list['atom_name']=='C12'].index]
Cc= atom_list[atom_list['atom_name']=='Cc'].index
comb=[(0,6),(1,7),(2,8),(3,9),(4,10),(5,11)]

for i,j in comb:
    for d1,d2,d3 in zip(Cc,Cm[i],Cm[j]):
        gaussian.addDonor(d1,d2,d3)

system.addForce(gaussian)
gaussian.getCutoffDistance()

gaussian.getNumDonors(),gaussian.getNumAcceptors()

print(len(system.getForces()))
print([f.usesPeriodicBoundaryConditions() for f in system.getForces()])


temperature=300*kelvin
integrator = VerletIntegrator(0.1*picoseconds)
simulation_platform = "OpenCL" # OpenCL, CUDA, CPU, or Reference
platform = Platform.getPlatformByName(simulation_platform)

simulation = Simulation(pdb.topology, system, integrator,platform)
simulation.context.setPositions(pdb.positions)
simulation.reporters.append(PDBReporter('output_%s.pdb'%simulation_platform, 1000),)
simulation.reporters.append(DCDReporter('output_%s.dcd'%simulation_platform, 1000),)
simulation.reporters.append(StateDataReporter('sim_%s.log'%simulation_platform, 1000, step=True,time=True,totalEnergy=True,
                                              kineticEnergy=True,potentialEnergy=True, temperature=True))
state = simulation.context.getState(getEnergy=True)
energy=state.getPotentialEnergy().value_in_unit(kilojoule_per_mole)
print (energy)
simulation.minimizeEnergy()
simulation.context.setVelocitiesToTemperature(300*kelvin)
simulation.step(10000000)

